@@ -13,11 +13,22 @@
 #ifndef _PARSER_UTILITIES_
 #define _PARSER_UTILITIES_
 
+
 #include "../link-grammar/link-includes.h"
 
-#define MAX_INPUT 2048
+#include "command-line.h"
+
+char *expand_homedir(const char *fn);
+void set_screen_width(Command_Options*);
+void initialize_screen_width(Command_Options *);
+int get_verbosity(void); // Defined in command-line.c
+
+#define MAX_INPUT_LINE 2048
 
 #ifdef _WIN32
+#ifndef mkdir
+#define mkdir(d, a) mkdir(d)
+#endif
 #ifndef __MINGW32__
 /* There is no ssize_t definition in native Windows. */
 #include <BaseTsd.h>
@@ -29,11 +40,21 @@ typedef SSIZE_T ssize_t;
 #define strncasecmp _strnicmp
 #endif
 
-char *get_console_line(void);
-void win32_set_utf8_output(void);
-char **argv2utf8(int);
+char **ms_windows_setup(int);
+int get_console_line(char *inbuf, int inbuf_size);
 int lg_isatty(int);
 #define isatty lg_isatty
+#else /* !_WIN32 */
+#define ms_windows_setup(argc) (argv)
 #endif /* _WIN32 */
 
+bool get_line(const char *, char **, unsigned int, FILE *, FILE *, bool);
+
+#if !HAVE_ASPRINTF
+int vasprintf(char ** restrict, const char * restrict, va_list)
+	GNUC_PRINTF(2,0);
+int asprintf(char ** restrict, const char * restrict, ...)
+	GNUC_PRINTF(2,3);
+#endif /* !HAVE_ASPRINTF */
+
 #endif // _PARSER_UTILITIES_
